# -*- coding: utf-8 -*-
"""A console interface based on curses. """

#
# Copyright (C) 2011 Anders Engström <ankan@ankan.eu>
#
# This file is part of pstorytime.
#
# pstorytime is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pstorytime is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pstorytime.  If not, see <http://www.gnu.org/licenses/>.

import traceback
import sys
import curses
import gobject
import glib
from os.path import isfile, isdir, join, expanduser, dirname, basename, exists, abspath
import signal
import os

import pygst
pygst.require("0.10")

# Don't touch my arguments!
argv = sys.argv
sys.argv = []
import gst
sys.argv = argv

import argparse
import time
import textwrap

from pstorytime.audiobook import AudioBook
from pstorytime.misc import PathGen, FileLock, DummyLock, LockedException
from pstorytime.misc import ns_to_str, parse_pos, LineStore, Struct, wrap_section
from pstorytime.timer import Timer
import pstorytime.audiobookargs
import pstorytime.version
import pstorytime.pandorabanner

class Select(object):
  """A window where a line is selected. It can either list all log entries or all files. Alternatively in can show the error log.
  """
  def __init__(self,conf,audiobook,controller,line):
    """Create window.

    Arguments:
      conf        Configuration struct. (Config file and arguments.)
      audiobook   Audiobook player object.
      controller  The object that handles user input.
      line        The line object which can show a title name.
    """
    self._window = curses.newwin(1,1,0,0)
    self._geom = Geometry.dummy()
    self._audiobook = audiobook
    self._controller = controller
    self._line = line
    self._line.set_title("Play Log")

    self._audiobook.connect("notify::playlog",self._on_playlog)

    self._last_entry = None

    self._logsel = LogSelect( self._window,
                              conf,
                              self._audiobook)
    self._filesel = FileSelect( self._window,
                                self._audiobook)
    self._errorsel = ErrorSelect( self._controller,
                                  self._window,
                                  self._audiobook)
    self._focus = self._logsel

    self._controller.connect("event",self._on_event)

  def get_geom(self):
    """Get current geometry of the window.
    """
    return self._geom

  def set_geom(self,geom):
    """Set geometry.

    Arguments:
      geom      The new geometry.
    """
    self._geom = geom
    self._logsel.set_geom(geom)
    self._filesel.set_geom(geom)
    self._errorsel.set_geom(geom)

    if self._geom.is_sane():
      self._window.resize(geom.h,geom.w)
      self._window.mvwin(geom.y,geom.x)
      self.update()

  def _on_playlog(self,ab,prop):
    """Callback when playlog is updated.

    Arguments:
      ab    Audiobook object.
      prop  The property that was changed.
    """
    if self._focus == self._logsel:
      playlog = self._audiobook.playlog
      if len(playlog)>0 and playlog[-1]!=self._last_entry:
        self.update()

  def _on_event(self,obj,event):
    """Callback when a user input event has been triggered.

    Arguments:
      obj       The conctroller object.
      event     The event that was triggered. (A line of text.)

    Returns:    True if the event was handled.
    """
    data = event.split()
    if len(data)>0:
      cmd = data[0]
      if cmd=="up":
        self._focus.move(-1)
        return True

      elif cmd=="down":
        self._focus.move(1)
        return True

      elif cmd=="ppage":
        self._focus.ppage()
        return True

      elif cmd=="npage":
        self._focus.npage()
        return True

      elif cmd=="begin":
        self._focus.move_to(0)
        return True

      elif cmd=="end":
        self._focus.move_to(None)
        return True

      elif cmd=="next_view":
        self._next_view()
        return True

      elif cmd=="playlog_view":
        self._playlog_view()
        return True

      elif cmd=="filelist_view":
        self._filelist_view()
        return True

      elif cmd=="errorlog_view":
        self._errorlog_view()
        return True

      elif cmd=="select":
        if len(data)==2:
          # There is an argument, parse it as a position in a file.
          (rel, pos) = parse_pos(data[1]) 
          # Make sure it is a valid position.
          if pos == None:
            return False
        else:
          # No argument.
          rel = None
          pos = None
        # Let the list view that is focused handle the event further.
        self._focus.select(rel,pos)
        return True

      return False

  def _next_view(self):
    """Change to the next view.
    """
    if self._focus == self._logsel:
      self._filelist_view()
    elif self._focus == self._filesel:
      self._errorlog_view()
    elif self._focus == self._errorsel:
      self._playlog_view()

  def _playlog_view(self):
    """Change to the playlog view.
    """
    self._errorsel.auto_update(False)
    self._focus = self._logsel
    self._line.set_title("Play Log")
    self.update()

  def _filelist_view(self):
    """Change to the filelist view.
    """
    self._errorsel.auto_update(False)
    self._focus = self._filesel
    self._line.set_title("File List")
    self.update()

  def _errorlog_view(self):
    """Change to the errorlog view.
    """
    self._errorsel.auto_update(True)
    self._focus = self._errorsel
    self._line.set_title("Error Log")
    self.update()

  def update(self):
    """Redraw this window.
    """
    self._focus.draw()

class LogSelect(object):
  """Display a list of log entries.
  """
  def __init__(self,window,conf,audiobook):
    """Create window.

    Arguments:
      window      The ncurses window to print to.
      conf        Configuration struct. (Config file and arguments.)
      audiobook   Audiobook player object.
    """
    self._conf = conf
    self._window = window
    self._geom = Geometry.dummy()
    self._audiobook = audiobook

    self._focus = None

  def get_geom(self):
    """Get current geometry of the window.
    """
    return self._geom

  def set_geom(self,geom):
    """Set geometry.

    Arguments:
      geom      The new geometry.
    """
    self._geom = geom

  def move(self,delta):
    """Move focus to another line.

    Arguments:
      delta     How many lines to move focus. Positive numbers to move the
                focus down, negative to move it up.
    """
    playlog = self._audiobook.playlog
    self._focus = calc_focus( length = len(playlog),
                              focus = self._focus,
                              delta = delta)
    self.draw()

  def ppage(self):
    """Move focus a page up.
    """
    playlog = self._audiobook.playlog
    num = min(self._geom.h, len(playlog))
    self._focus = calc_ppage( length = len(playlog),
                              num = num,
                              focus = self._focus)
    self.draw()

  def npage(self):
    """Move focus a page down.
    """
    playlog = self._audiobook.playlog
    num = min(self._geom.h, len(playlog))
    self._focus = calc_npage( length = len(playlog),
                              num = num,
                              focus = self._focus)
    self.draw()

  def move_to(self,focus):
    """Move focus to an absolute index.
    
    Arguments:
      focus     The line number to focus. If None, no line is selected.
    """
    playlog = self._audiobook.playlog
    self._focus = calc_focus( length = len(playlog),
                              focus = focus)
    self.draw()

  def select(self,rel,bufpos):
    """Select the current entry.

    Arguments:
      rel       True if bufpos is a relative position.
      bufpos    A stream position. (Possibly None.)
    """
    ab = self._audiobook
    playlog = ab.playlog
    if self._focus == None:
      # No line is selected, do a seek according to position information.
      if rel:
        ab.dseek(bufpos)
      elif rel!=None:
        ab.seek(None,bufpos)
    elif self._focus <= len(playlog):
      # A line is selected. Seek to another file.
      filename = playlog[self._focus].filename
      position = playlog[self._focus].position
      if rel:
        # Relative seeks are relative to log entry position.
        ab.seek(filename,position+bufpos)
      elif rel!=None:
        # User has entered an absolute position on the input buffer.
        # Seek to that position in the selected file.
        ab.seek(filename,bufpos)
      else:
        # No position given on the input buffer. Seek to position of log entry.
        ab.seek(filename,position)

  def draw(self):
    """Draw the current state.
    """

    if self._geom.is_sane():
      self._window.erase()

      playlog = self._audiobook.playlog

      # Make sure the focus index is valid.
      self._focus = calc_focus( length = len(playlog),
                                focus = self._focus)

      # Number of entries to show.
      num = min(self._geom.h, len(playlog))

      focus = self._focus

      # Compute index of first log entry to show.
      if focus==None:
        start = max(0, len(playlog)-num)
      else:
        start = max(0, min(len(playlog)-num, focus - num/2))

      for i in xrange(0, num):
        # Position in playlog
        logi = i + start
        # Check if this is the currently selected line
        if logi == focus:
          mark = "-> "
          attr = curses.A_REVERSE | curses.A_BOLD
        else:
          mark = "   "
          attr = curses.A_NORMAL

        walltime = time.strftime( "%Y-%m-%d %H:%M:%S",
                                  time.localtime(playlog[logi].walltime))

        # Generate a string of fixed length that contains the event name.
        event = playlog[logi].event[:self._conf.event_len]
        event += " " * (self._conf.event_len - len(event))

        # Format all stuff before filename
        part0 = "{mark}{walltime} {event} ".format(
          mark = mark,
          walltime = walltime,
          event = event)

        # Format all stuff after filename
        position = ns_to_str(playlog[logi].position)
        duration = ns_to_str(playlog[logi].duration)

        part2 = " {position} / {duration}".format(
          position=position,
          duration=duration)

        # Compute maximum length of filename
        part1len = max(0, self._geom.w - 1 - len(part0) - len(part2))
        # Take the end of filename, if it is too long.
        part1 = playlog[logi].filename[-part1len:]

        # Combine into complete line.
        pad = " " * (part1len - len(part1))
        line = part0 + part1 + pad + part2

        self._window.addnstr(i, 0, line, self._geom.w-1,attr)
      self._window.refresh()

class FileSelect(object):
  """Display a list of filenames.
  """
  def __init__(self,window,audiobook):
    """Create window.

    Arguments:
      window      The ncurses window to print to.
      audiobook   Audiobook player object.
    """
    self._window = window
    self._geom = Geometry.dummy()
    self._audiobook = audiobook

    self._focus = None

  def get_geom(self):
    """Get current geometry of the window.
    """
    return self._geom

  def set_geom(self,geom):
    """Set geometry.

    Arguments:
      geom      The new geometry.
    """
    self._geom = geom

  def move(self,delta):
    """Move focus to another line.

    Arguments:
      delta     How many lines to move focus. Positive numbers to move the
                focus down, negative to move it up.
    """
    filelist = self._audiobook.list_files()
    self._focus = calc_focus( length = len(filelist),
                              focus = self._focus,
                              delta = delta)
    self.draw()

  def ppage(self):
    """Move focus a page up.
    """
    filelist = self._audiobook.list_files()
    num = min(self._geom.h, len(filelist))
    self._focus = calc_ppage( length = len(filelist),
                              num = num,
                              focus = self._focus)
    self.draw()

  def npage(self):
    """Move focus a page down.
    """
    filelist = self._audiobook.list_files()
    num = min(self._geom.h, len(filelist))
    self._focus = calc_npage( length = len(filelist),
                              num = num,
                              focus = self._focus)
    self.draw()

  def move_to(self,focus):
    """Move focus to an absolute index.
    
    Arguments:
      focus     The line number to focus. If None, no line is selected.
    """
    filelist = self._audiobook.list_files()
    self._focus = calc_focus( length = len(filelist),
                              focus = focus)
    self.draw()

  def select(self,rel,bufpos):
    """Select the current entry.

    Arguments:
      rel       True if bufpos is a relative position.
      bufpos    A stream position. (Possibly None.)
    """
    ab = self._audiobook
    filelist = ab.list_files()
    if self._focus == None:
      # No line selected. Do a simple seek.
      if rel:
        ab.dseek(bufpos)
      elif rel!=None:
        ab.seek(None,bufpos)
    elif 0 <= self._focus < len(filelist):
      # Line selected, seek to that file and the position given on the buffer
      # line. (Negative position indicates an earlier file.)
      ab.seek(filelist[self._focus],bufpos)

  def draw(self):
    """Draw the current state.
    """
    if self._geom.is_sane():
      self._window.erase()

      filelist = self._audiobook.list_files()

      # Make sure the focus index is valid.
      self._focus = calc_focus( length = len(filelist),
                                focus = self._focus)

      # Number of entries to show.
      num = min(self._geom.h, len(filelist))

      focus = self._focus

      # Compute index of first log entry to show.
      if focus==None:
        start = max(0, len(filelist)-num)
      else:
        start = max(0, min(len(filelist)-num, focus - num/2))

      for i in xrange(0, num):
        # Position in filelist
        listi = i + start
        # Check if this is the currently selected line
        if listi == focus:
          mark = "-> "
          attr = curses.A_REVERSE | curses.A_BOLD
        else:
          mark = "   "
          attr = curses.A_NORMAL

        # Compute maximum length of filename
        part0len = max(0, self._geom.w - 1 - len(mark))
        # Take the end of filename, if it is too long.
        part0 = filelist[listi][-part0len:]

        # Combine into complete line.
        pad = " " * (part0len - len(part0))
        line = mark + part0 + pad

        self._window.addnstr(i, 0, line, self._geom.w-1,attr)
      self._window.refresh()

class ErrorSelect(object):
  """List all error log entries.
  """
  def __init__(self,controller,window,audiobook):
    """Create window.

    Arguments:
      controller  The object that handles user input.
      window      The ncurses window to print to.
      audiobook   Audiobook player object.
    """
    self._controller = controller
    self._window = window
    self._geom = Geometry.dummy()
    self._audiobook = audiobook

    self._auto_update = False

    self._error_log = []
    self._controller.connect("error",self._on_error)
    self._audiobook.connect("error",self._on_error)

    self._focus = None

  def get_geom(self):
    """Get current geometry of the window.
    """
    return self._geom

  def set_geom(self,geom):
    """Set geometry.

    Arguments:
      geom      The new geometry.
    """
    self._geom = geom

  def _on_error(self,obj,msg):
    """Callback for when errors occur.

    Arguments:
      obj       The object that the error came from.
      msg       The error message.
    """
    walltime = time.strftime("%Y-%m-%d %H:%M:%S")
    self._error_log.append((walltime, msg))
    if self._auto_update:
      self.draw()

  def move(self,delta):
    """Move focus to another line.

    Arguments:
      delta     How many entries to move focus. Positive numbers to move the
                focus down, negative to move it up.
    """
    self._focus = calc_focus( length = len(self._error_log),
                              focus = self._focus,
                              delta = delta)
    self.draw()

  def ppage(self):
    """Move focus a page up.
    """
    self.move(-1)

  def npage(self):
    """Move focus a page down.
    """
    self.move(1)

  def move_to(self,focus):
    """Move focus to an absolute index.
    
    Arguments:
      focus     The number of the log entry to focus. If None, no entry is selected.
    """
    self._focus = calc_focus( length = len(self._error_log),
                              focus = focus)
    self.draw()

  def select(self,rel,bufpos):
    """Seek to the given position. (Nothing to select in the error log.)

    Arguments:
      rel       True if bufpos is a relative position.
      bufpos    A stream position. (Possibly None.)
    """
    ab = self._audiobook
    if rel:
      ab.dseek(bufpos)
    elif rel!=None:
      ab.seek(None,bufpos)

  def auto_update(self,state):
    """Control if the list should be automatically updated.
    
    Arguments:
      state     If the list should be automatically updated.
    """
    self._auto_update = state

  def draw(self):
    """Draw the current state.
    """
    if self._geom.is_sane():
      self._window.erase()

      sel_attr = curses.A_REVERSE | curses.A_BOLD

      # Make sure the focus index is valid.
      self._focus = calc_focus( length = len(self._error_log),
                                focus = self._focus)

      log = self._error_log
      f = self._focus
      h = self._geom.h
      w = self._geom.w

      if f == None:
        # No entry is selected.
        sel = []

        # Worst case number of entries required to fill in each direction.
        num_fill = h

        before = []
        
        # Generate a list of all entries we may wish to display before the selection.
        for (header,body) in log[-num_fill:]:
          before += wrap_section(header,body,w-4)

        # No entry is selected (we are below the last line)
        # so nothing to display after it.
        after = []
      else:
        # A line is selected.

        # Format the lines of the selected entry.
        sel = wrap_section( log[f][0],
                            log[f][1],
                            w-4)

        # Worst case number of entries required to fill in each direction.
        num_fill = h - len(sel)

        # Generate a list of all entries we may wish to display before the selection.
        before = []
        for (header,body) in log[ max(0,f-num_fill):f ]:
          before += wrap_section(header,body,w-4)

        # And after...
        after = []
        for (header,body) in log[ f+1:f+1+num_fill ]:
          after += wrap_section(header,body,w-4)
        
      # Find out how many lines to display before the selection, if we are somewhere
      # in the middle of the error log. (Neither the beginning nor the end.)
      nbefore = min(h/2 - len(sel)/2, len(before))

      # From this we can find out how many lines to display after the selection.
      nafter = min(h - nbefore - len(sel), len(after))

      # Now we can find out how many lines to actually display before.
      nbefore += h - nbefore - len(sel) - nafter


      i = 0
      for line in before[-nbefore:]:
        self._window.addnstr(i, 0, '   '+line, w-1)
        i += 1

      for line in sel[0:1]:
        self._window.addnstr(i, 0, '-> '+line, w-1, sel_attr)
        i += 1

      for line in sel[1:h]:
        self._window.addnstr(i, 0, '   '+line, w-1, sel_attr)
        i += 1

      for line in after[:nafter]:
        self._window.addnstr(i, 0, '   '+line, w-1)
        i += 1

      self._window.refresh()

def calc_focus(length,focus,delta=0):
  """Calculate a new focus index.

  Arguments:
    length    Number of entries to choose from.
    focus     Base focus. (None, if no line is selected.)
    delta     Offset from base focus.
  """
  if focus == None:  
    if delta>0:
      focus = delta-1
    else:
      focus = delta+length
  else:
    focus += delta
  
  # If we end up outside the list, no line is selected.
  if focus<0 or focus>=length:
    focus = None

  return focus

def calc_ppage(length,num,focus):
  """Compute focus index after a page up.

  Arguments:
    length    Number of entries to choose from.
    num       Number of entries that are displayed.
    focus     Index of currently focused entry. (Possibly None.)
  """
  # Special case for all entries at the bottom of the list.
  if focus == None or focus > length-num/2:
    focus = length-num/2

  return max(0, focus - num)

def calc_npage(length,num,focus):
  """Compute focus index after a page down.

  Arguments:
    length    Number of entries to choose from.
    num       Number of entries that are displayed.
    focus     Index of currently focused entry. (Possibly None.)
  """
  # Page up has no effect when no line is selected.
  if focus == None:
    return None

  # Special case for all entries at the top of the list.
  if focus < num/2:
    focus = num/2

  # Compute new focus position.
  newfocus = focus + num
  
  # If it is outside the list of entries, do not select any line.
  if newfocus > length:
    return None
  else:
    return newfocus

class Controller(gobject.GObject):
  """Handle user input.
  """
  HEIGHT=1

  __gsignals__ = {
    'error' : ( gobject.SIGNAL_RUN_LAST,
                gobject.TYPE_NONE,
                (gobject.TYPE_STRING,)),
    'event' : ( gobject.SIGNAL_RUN_LAST,
              gobject.TYPE_BOOLEAN,
              (gobject.TYPE_STRING,),
              gobject.signal_accumulator_true_handled)
  }

  def __init__(self,charmap,keytime):
    """Create window.

    Arguments:
      charmap     Dictionary containing mappings from key names to events.
      keytime     How long to display keynames after keypress.
    """
    gobject.GObject.__init__(self)
    self._keytime = keytime

    self._geom = Geometry.dummy()
    self._window = curses.newwin(1,1,0,0)
    self._window.nodelay(1)
    self._window.keypad(1)
    curses.raw()

    self._charmap = charmap

    self._key = None
    self._clear_timer = Timer(self._keytime,self._clear_key,repeat=False)
    self._buffer = ""
    self.connect("event",self._on_event)

  def poll(self):
    """Read all available user input in non-blocking mode until there is no more to read.
    """
    while True:
      # Read characters and process them until there are no more pending ones.
      ch = self._window.getch()
      if ch == -1:
        break

      try:
        key = curses.keyname(ch)

        if self._keytime>0 and key != "KEY_RESIZE":
          self._key = key
          self.update()
          self._clear_timer.start()

        event = self._charmap[key]
        self.broadcast(event)
      except (KeyError, IndexError):
        pass

  def broadcast(self,command):
    """Broadcast an event so that different parts of the program can react to it.

    Arguments:
      command     The command to execute.
    """
    for event in command.split(';'):
      # Replace {b} with the string in the input buffer.
      event_formatted = event.strip().format(b=self._buffer).strip()
      # Ignore the empty string.
      if len(event_formatted)>0:
        # Emit event and if it is not handled anywhere emit an error.
        if not self.emit("event",event_formatted):
          self.emit('error','Invalid event: "{0}"'.format(event_formatted))
          # Don't handle the rest of the events.
          break

  def _on_event(self,obj,event):
    """Callback to handle buffer events.

    Arguments:
      obj       The controller.
      event     The string describing what to do.

    Returns:    True if the event is handled.
    """
    data = event.split()
    if len(data)>0 and data[0]=="buffer":
      if data[1] == "store":
        self._buffer += data[2]
        self.update()
        return True
      elif data[1] == "erase":
        self._buffer = self._buffer[:-1]
        self.update()
        return True
      elif data[1] == "clear":
        self._buffer = ""
        self.update()
        return True

    return False

  def _clear_key(self):
    """Callback for when the last key pressed should be hidden.
    """
    self._key = None
    self.update()

  def get_geom(self):
    """Get current geometry of the window.
    """
    return self._geom

  def set_geom(self,geom):
    """Set geometry.

    Arguments:
      geom      The new geometry.
    """
    self._geom = geom
    if self._geom.is_sane():
      self._window.resize(geom.h,geom.w)
      self._window.mvwin(geom.y,geom.x)
      self.update()

  def update(self):
    """Redraw this window.
    """
    if self._geom.is_sane():
      self._window.erase()

      # Always there.
      prefix = "> "
      
      # The name of the last key pressed is shown last on the line.
      if self._key!=None:
        keystr = " ({0})".format(self._key)
      else:
        keystr = ""

      # Maximum number of characters of the buffer that can be shown.
      maxbuf = self._geom.w - len(keystr) - len(prefix) - 1
      # Take the end of the buffer and show as many characters as possible.
      bufstr = self._buffer[-maxbuf:]
      # Place spaces between end of the buffer line and the last key pressed information.
      spacing = " "*(maxbuf-len(bufstr))
      
      line = prefix+bufstr+spacing+keystr
      self._window.addnstr(0,0,line,self._geom.w-1)

      self._window.refresh()

class FifoReader(object):
  """Read commands from a fifo.
  """
  def __init__(self,cmdpipe,broadcast):
    """Create window.

    Arguments:
      cmdpipe     Path to a fifo to read commands from. (None to disable reading.)
      broadcast   A function taking a string as argument that is called for
                  each line that is read.
    """
    self._cmdpipe = cmdpipe
    self._broadcast = broadcast
    self._buffer = ""
    self._f = None
    self._event_source = None
    if cmdpipe!=None and cmdpipe!="":
      self._open()

  def _open(self):
    """Open the fifo for reading.
    """
    # Prepare a fifo.
    dirpath = dirname(self._cmdpipe)
    if not isdir(dirpath) and dirpath!='':
      os.makedirs(dirpath)
    if not exists(self._cmdpipe):
      os.mkfifo(self._cmdpipe)

    # Open fifo in non-blocking read mode.
    self._f = os.open(self._cmdpipe, os.O_RDONLY | os.O_NONBLOCK)
    # Register callback for when something has happened.
    self._event_source = glib.io_add_watch( self._f,
                                            glib.IO_IN | glib.IO_ERR | glib.IO_HUP,
                                            self._io)

  def _io(self,source,condition):
    """Callback for when something has happened with the fifo.

    Arguments:
      source      The event source that triggered the callback.
      condition   What has happened with the file handler.
    """
    if condition == glib.IO_IN:
      # Ready to read.
      try:
        # Non-blocking read.
        self._buffer += os.read(self._f, 200)

        # Get all completed lines.
        (prefix,sep,self._buffer) = self._buffer.rpartition("\n")
        # Broadcast these lines.
        for line in prefix.splitlines():
          self._broadcast(line)
        return True
      except (OSError, IOError):
        # If the reading fails, reopen the file.
        self._reopen()
        return False
    else:
      # Otherwise we cannot read from this fifo any more, reopen it.
      self._reopen()
      return False

  def _close(self):
    """Close the file handler for the fifo.
    """
    if self._event_source != None:
      glib.source_remove(self._event_source)
    if self._f != None:
      os.close(self._f)

  def _reopen(self):
    """Reopen the file handler.
    """
    self._close()
    self._open()

  def quit(self):
    """Shut down the fifo reader.
    """
    self._close()

class DirName(object):
  """Display the name of the directary that is playing.
  """

  HEIGHT=1
  ATTR = curses.A_NORMAL

  def __init__(self,directory):
    """Create window.

    Arguments:
      directory     Name of audiobook directory.
    """
    self._directory = directory
    self._window = curses.newwin(1,1,0,0)
    self._geom = Geometry.dummy()

  def get_geom(self):
    """Get current geometry of the window.
    """
    return self._geom

  def set_geom(self,geom):
    """Set geometry.

    Arguments:
      geom      The new geometry.
    """
    self._geom = geom
    if self._geom.is_sane():
      self._window.resize(geom.h,geom.w)
      self._window.mvwin(geom.y,geom.x)
      self.update()

  def update(self):
    """Redraw this window.
    """
    if self._geom.is_sane():
      maxlen = self._geom.w - 1
      line = self._directory[-maxlen:]

      self._window.erase()
      self._window.addnstr(0,0,line,maxlen,self.ATTR)
      self._window.refresh()

class HLine(object):
  """Display a horizontal line.
  """

  HEIGHT=1
  CHARS_BEFORE_TITLE=2
  FORMAT="[ {title} ]"

  def __init__(self,title=None):
    """Create window.

    Arguments:
      title     An optional title to display on the line.
    """
    self._title = title
    self._window = curses.newwin(1,1,0,0)
    self._geom = Geometry.dummy()

  def get_geom(self):
    """Get current geometry of the window.
    """
    return self._geom

  def set_geom(self,geom):
    """Set geometry.

    Arguments:
      geom      The new geometry.
    """
    self._geom = geom
    if self._geom.is_sane():
      self._window.resize(geom.h,geom.w)
      self._window.mvwin(geom.y,geom.x)
      self.update()

  def set_title(self,title):
    """Update line title.

    Arguments:
      title     The new title.
    """
    self._title = title
    self.update()

  def get_title(self,title):
    """Get line title.

    Returns:  The current title.
    """
    return self._title

  def update(self):
    """Redraw this window.
    """
    before = HLine.CHARS_BEFORE_TITLE
    line = HLine.FORMAT
    if self._geom.is_sane():
      self._window.erase()

      if self._title == None or self._geom.w<=before+1:
        self._window.hline(0,0,curses.ACS_HLINE,self._geom.w)
      else:
        self._window.hline(0,0,curses.ACS_HLINE,before)
        line = line.format(title=self._title)
        self._window.addnstr(0,before,line,self._geom.w-before-1)
        remaining = self._geom.w-before-len(line)
        if remaining>0:
          self._window.hline(0,before+len(line),curses.ACS_HLINE,remaining)
      self._window.refresh()

class About(object):
  """Display the program name and version.
  """

  ABOUT = "pstorytime {ver}".format(ver=pstorytime.version.version)
  WIDTH = len(ABOUT)+2
  HEIGHT=1
  ATTR = curses.A_NORMAL

  def __init__(self):
    """Create window.
    """
    self._window = curses.newwin(1,1,0,0)
    self._geom = Geometry.dummy()

  def get_geom(self):
    """Get current geometry of the window.
    """
    return self._geom

  def set_geom(self,geom):
    """Set geometry.

    Arguments:
      geom      The new geometry.
    """
    self._geom = geom
    if self._geom.is_sane():
      self._window.resize(geom.h,geom.w)
      self._window.mvwin(geom.y,geom.x)
      self.update()

  def update(self):
    """Redraw this window.
    """
    if self._geom.is_sane():
      self._window.erase()
      if self._geom.w>=2:
        self._window.addnstr(0,1,self.ABOUT,self._geom.w-2,self.ATTR)
      self._window.refresh()

class Volume(object):
  """Display playback volume.
  """
  HEIGHT = 2
  WIDTH = 6

  def __init__(self,audiobook):
    """Create window.

    Arguments:
      audiobook   Audiobook player object.
    """
    self._gstvol = audiobook.gst().get_by_name('volume-control')
    if self._gstvol != None:
      self._gstvol.connect("notify::volume",self._on_volume)
    self._window = curses.newwin(1,1,0,0)
    self._geom = Geometry.dummy()

  def get_geom(self):
    """Get current geometry of the window.
    """
    return self._geom

  def set_geom(self,geom):
    """Set geometry.

    Arguments:
      geom      The new geometry.
    """
    self._geom = geom
    if self._geom.is_sane():
      self._window.resize(geom.h,geom.w)
      self._window.mvwin(geom.y,geom.x)
      self.update()

  def _on_volume(self,obj,prop):
    """Callback for when the volume has changed in gstreamer.

    Arguments:
      obj       Gstreamer object.
      prop      The property that has changed.
    """
    self.update()

  def update(self):
    """Redraw this window.
    """
    if self._geom.is_sane():
      if self._gstvol != None:
        try:
          vol = "{0:.0%}".format(self._gstvol.get_property("volume"))
        except gst.QueryError:
          vol = "--%"
      else:
        vol = "--%"
      self._window.erase()
      if self._geom.h>=1:
        self._window.addnstr(0,0,"  Vol",self._geom.w-1)
      if self._geom.h>=2:
        self._window.addnstr(1,0,"{0:>5}".format(vol),self._geom.w-1)
      self._window.refresh()

class Status(object):
  """Display general playback status. (State, filename, position, etc.)
  """
  HEIGHT = 2

  def __init__(self,conf,audiobook,stoptimer,interval):
    """Create the audiobook view.
    
    Arguments:
      conf        The parsed program configuration.
      audiobook   The audiobook object to create a view for.
      stoptimer   The stop timer to show state of.
      interval    How often to show position while playing.
    """
    self._audiobook = audiobook
    self._stoptimer = stoptimer

    self._window = curses.newwin(1,1,0,0)
    self._geom = Geometry.dummy()

    self._audiobook.connect("position",self._on_change)
    self._audiobook.connect("notify::playing",self._on_playing)
    self._stoptimer.connect("changed",self._on_change)
    self._update_timer = Timer(interval*1000, self._on_update_timer, repeat=True)

  def get_geom(self):
    """Get current geometry of the window.
    """
    return self._geom

  def set_geom(self,geom):
    """Set geometry.

    Arguments:
      geom      The new geometry.
    """
    self._geom = geom
    if self._geom.is_sane():
      self._window.resize(geom.h,geom.w)
      self._window.mvwin(geom.y,geom.x)
      self.update()

  def _on_playing(self,ab,prop):
    """Playing state updated.
    
    Arguments:
      ab    The audiobook that this is a view for.
      prop  The property that was updated.
    """
    if ab.playing and (not self._update_timer.started()):
      self._update_timer.start()
    if (not ab.playing) and self._update_timer.started():
      self._update_timer.stop()

  def _on_change(self,obj):
    """A hint has been received that it is a good idea to update the state.
    
    Arguments:
      obj    The object that the change is associated with.
    """
    self.update()

  def _on_update_timer(self):
    """Callback for when it is time to update the playback information.
    """
    self.update()

  def update(self):
    """Redraw this window.
    """
    if self._geom.is_sane():
      (filename,position,duration) = self._audiobook.position()
      if filename == None:
        filename = ""

      if self._audiobook.playing:
        state = "Playing"
      elif self._audiobook.eob:
        state = "End"
      else:
        state = "Paused"

      # Format first line.
      prefix = "File: "
      maxchars = self._geom.w - len(prefix) - 1
      first = prefix + filename[-maxchars:]

      # Format stoptimer state
      if self._stoptimer.show:
        stoptimer = "(Timer: {stoptime}{stopped})".format(
            stoptime = time.strftime(
              "%H:%M:%S",
              time.localtime(self._stoptimer.time)),
          stopped = "" if self._stoptimer.running else " Stopped")
      else:
        stoptimer = ""

      # Format second line.
      second = "{position} / {duration} [{state}] {stoptimer}".format(
        position=ns_to_str(position),
        duration=ns_to_str(duration),
        state=state,
        stoptimer=stoptimer)

      self._window.erase()
      if self._geom.h>=2:
        self._window.addnstr(0,0,first,self._geom.w-1)
        self._window.addnstr(1,0,second,self._geom.w-1)
      else:
        self._window.addnstr(0,0,second,self._geom.w-1)
      self._window.refresh()

class StopTimer(gobject.GObject):
  __gsignals__ = {
    'changed' : ( gobject.SIGNAL_RUN_LAST,
                  gobject.TYPE_NONE,
                  tuple())
  }

  @property
  def time(self):
    """The time from midnight to stop at (in ns.) """
    return self._time

  @property
  def running(self):
    """If the timer is running. """
    return self._running

  @property
  def show(self):
    """If the timer should be shown. """
    return self._show

  def __init__(self,controller,audiobook):
    """Create the stop timer handler.

    Arguments:
      controller  The object that handles user input.
      audiobook   Audiobook player object.
    """
    gobject.GObject.__init__(self)
    self._controller = controller
    self._audiobook = audiobook

    self._controller.connect("event",self._on_event)

    self._time = 0
    self._running = False
    self._show = False

    self._timer_id = None

  def _on_event(self,obj,event):
    """Handle user input events.

    Arguments:
      obj       The controller this event came from.
      event     The string containing the command to evaluate.

    Returns:    True if the event was handled.
    """
    data = event.split()

    if len(data)>0:
      if data[0]=="timer":
        if 1 <= len(data) <= 2 and self._timer_id != None:
          glib.source_remove(self._timer_id)
          self._timer_id = None

        if len(data)==1:
          # No time given, stop timer.
          self._running = False
          self._show = False
          self.emit("changed")
          return True
        elif len(data)==2:
          # Time given, start timer.
          (rel, stoptime) = parse_pos(data[1])
          if stoptime == None:
            # Failed to parse time.
            return False
          stoptime = stoptime/1000**3
          curtime = int(time.time())
          if rel:
            stoptime += curtime
          else:
            # Absolute times are actually relative to the start of the first day
            # where that time of day is in the future.
            t = time.localtime(curtime)
            stoptime += curtime - t.tm_hour*60**2 - t.tm_min*60 - t.tm_sec
            while stoptime<curtime:
              stoptime += 24*60**2
          self._timer_id = glib.timeout_add_seconds(stoptime-curtime,self._on_timer)
          self._time = stoptime
          self._show = True
          self._running = True
          self.emit("changed")
          return True

    return False

  def _on_timer(self):
    """Timer has fired, pause playback.
    """
    self._audiobook.pause()
    self._running = False
    self.emit("changed")
    return False

class ErrorBox(object):
  """Display the most recent error messages.
  """
  def __init__(self,controller,audiobook,errortime):
    """Create window.

    Arguments:
      controller  The object that handles user input.
      audiobook   Audiobook player object.
      errortime   How long a error message is considered hot.
    """
    self._controller = controller
    self._audiobook = audiobook
    self._window = curses.newwin(1,1,0,0)
    self._geom = Geometry.dummy()
    self._errortime = errortime

    self._error_msg = ""
    self._error_walltime = ""

    self._controller.connect("error",self._on_error)
    self._audiobook.connect("error",self._on_error)

    self._hot = False
    self._cool_timer = Timer(self._errortime,self._cool,repeat=False)

  def _cool(self):
    """Callback when the cool timer for errors has triggered.
    """
    self._hot = False
    self._controller.broadcast("resize")

  def get_geom(self):
    """Get current geometry of the window.
    """
    return self._geom

  def set_geom(self,geom):
    """Set geometry.

    Arguments:
      geom      The new geometry.
    """
    self._geom = geom
    if self._geom.is_sane():
      self._window.resize(geom.h,geom.w)
      self._window.mvwin(geom.y,geom.x)
      self.update()

  def _on_error(self,obj,msg):
    """Callback for when errors occur.

    Arguments:
      obj       The object that the error came from.
      msg       The error message.
    """
    self._error_walltime = time.strftime("%Y-%m-%d %H:%M:%S")
    self._error_msg = msg
    self._hot = True
    if self._errortime>=0:
      self._cool_timer.start()
    self._controller.broadcast("resize")

  def lines(self,width):
    if self._hot:
      return len(wrap_section(self._error_walltime,
                              self._error_msg,
                              width-1))
    else:
      return 0

  def update(self):
    """Redraw this window.
    """
    if self._geom.is_sane():
      self._window.erase()

      if self._hot:
        lines = wrap_section( self._error_walltime,
                              self._error_msg,
                              self._geom.w-1)
      else:
        lines = []

      i = 0
      for line in lines[:self._geom.h]:
        self._window.addnstr(i,0,line,self._geom.w-1)
        i += 1

      self._window.refresh()

class GeometryFactory(object):
  """Build window geometries based on top geometry.
  """
  def __init__(self,topgeom):
    """Create the geometry factory.

    Arguments:
      topgeom   The top geometry to use.
    """
    self._topgeom = topgeom

  def build(self,h,w,y,x):
    """Build size and position storage object.

    The specified size is shrunk so that it fits within the top geometry.

    Positions will also always be made greater than or equal to 0.

    Arguments:
      h     Height.
      w     Width.
      y     Line number from top.
      x     Column from left.
    """
    lim = self._topgeom 

    oy = y
    ox = x

    y = max(lim.y, y)
    x = max(lim.x, x)

    h = max(0, min(lim.h, h+y-oy))
    w = max(0, min(lim.w, w+x-ox))

    return Geometry(h = h,
                    w = w,
                    y = y,
                    x = x)

  def build_above(self,h,w,y,x):
    """Build size and position storage object, with the y coordinate given as
    the line after the end of the box.

    The specified size is shrunk so that it fits within the top geometry.

    Positions will also always be made greater than or equal to 0.

    Arguments:
      h     Height.
      w     Width.
      y     Line number from top.
      x     Column from left.
    """
    return self.build(h = h,
                      w = w,
                      y = y-h,
                      x = x)


class Geometry(object):
  """Group height, width, line and column together and provide some abstraction.
  """
  @staticmethod
  def fromWindow(window):
    """Create a geometry with the same size and position as the given ncurses window.

    Arguments:
      window    The ncurses window to get size and position from.
    """
    return Geometry.fromSizePos(window.getmaxyx(),window.getbegyx())

  @staticmethod
  def fromSizePos(size,pos):
    """Create a geometry from a size and a position tuple.
    """
    return Geometry(size[0],size[1],pos[0],pos[1])

  @staticmethod
  def dummy():
    """Create a dummy geometry.

    Returns:  An geometry with zero height and width, that is positioned at origo.
    """
    return Geometry(0,0,0,0)

  def __init__(self,h,w,y,x):
    """Create size and position storage object.

    Arguments:
      h     Height.
      w     Width.
      y     Line number from top.
      x     Column from left.
    """
    self.h = h
    self.w = w
    self.y = y
    self.x = x

  def is_sane(self):
    """Check if the geometry is sane. That is, it is the position is positive
    and the window can show atleast one character.
    """
    return self.h>0 and self.w>0 and self.y>=0 and self.x>=0

  def __str__(self):
    """Create a string representation of the geometry.
    """
    return "({g.h},{g.w},{g.y},{g.x})".format(g=self)

class CursesUI(object):
  """A console audiobook player interface that uses ncurses.
  """
  def __init__(self,conf,directory,stdscr):
    """Create window.

    Arguments:
      conf        Configuration struct. (Config file and arguments.)
      directory   The directory that the audiobook player should load files from.
      stdscr      The root ncurses window.
    """
    self._conf = conf
    self._window = stdscr
    self._audiobook = AudioBook(conf,directory)

    self._quit = False

    try:
      curses.curs_set(0)
    except curses.error:
      pass

    gobject.threads_init()
    self._gloop = glib.MainLoop()
    self._gcontext = self._gloop.get_context()

    if conf.default_bindings:
      charmap = { "^L":"redraw",
                  "^C":"quit",
                  "q":"quit",
                  " ":"play_pause",
                  "KEY_LEFT":"seek -10",
                  "h":"seek -10",
                  "KEY_RIGHT":"seek +10",
                  "l":"seek +10",
                  "u":"volume +0.1",
                  "d":"volume -0.1",
                  "k":"up",
                  "j":"down",
                  "KEY_UP":"up",
                  "KEY_DOWN":"down",
                  "KEY_HOME":"begin",
                  "H":"begin",
                  "KEY_END":"end",
                  "F":"end",
                  "KEY_PPAGE":"ppage",
                  "KEY_NPAGE":"npage",
                  "^I":"next_view",
                  "p":"playlog_view",
                  "f":"filelist_view",
                  "e":"errorlog_view",
                  "*":"mark *",
                  "^J":"select {b}",
                  "t":"timer {b}; play",
                  "T":"timer",
                  "1":"buffer store 1",
                  "2":"buffer store 2",
                  "3":"buffer store 3",
                  "4":"buffer store 4",
                  "5":"buffer store 5",
                  "6":"buffer store 6",
                  "7":"buffer store 7",
                  "8":"buffer store 8",
                  "9":"buffer store 9",
                  "0":"buffer store 0",
                  ":":"buffer store :",
                  "+":"buffer store +",
                  "-":"buffer store -",
                  "=":"buffer clear",
                  "KEY_BACKSPACE":"buffer erase",
                  "^?":"buffer erase"}
    else:
      charmap = {}

    for binding in conf.bind:
      if len(binding)==2:
        charmap[binding[0]] = binding[1]
      elif len(binding)==1:
        del charmap[binding[0]]
                
    charmap["KEY_RESIZE"] = "resize"
    self._controller = Controller(charmap=charmap,
                                  keytime=conf.keytime)
    self._controller.connect("event",self._on_event)

    self._errorbox = ErrorBox(controller=self._controller,
                              audiobook=self._audiobook,
                              errortime=conf.errortime)

    self._errorline = HLine(title="Error")

    self._fiforeader = FifoReader(cmdpipe=conf.cmdpipe,
                                  broadcast=self._controller.broadcast)

    self._actuator = Actuator(conf=conf,
                              audiobook=self._audiobook,
                              controller=self._controller)

    self._volume = Volume(audiobook=self._audiobook)

    self._stoptimer = StopTimer(audiobook=self._audiobook,
                                controller=self._controller)

    self._status = Status(conf=conf,
                          audiobook=self._audiobook,
                          stoptimer=self._stoptimer,
                          interval=1)

    self._dirname = DirName(directory=directory)

    self._about = About()

    self._selectline = HLine()

    self._select = Select(conf=conf,
                          audiobook=self._audiobook,
                          controller=self._controller,
                          line=self._selectline)

    gobject.io_add_watch(sys.stdin, gobject.IO_IN, lambda src,cond: True)
    gobject.io_add_watch(sys.stdin, gobject.IO_ERR|gobject.IO_HUP, self._on_stdin_fail)

    self._update_layout()
  
  def _on_stdin_fail(self,src,cond):
    """Callback for when stdin is closed or otherwise fails.

    Arguments:
      src       The event source that triggered the callback.
      cond      The condition that caused the callback.
    """
    self.quit()
    return True

  def get_geom(self):
    """Get current geometry of the window.
    """
    geom = Geometry.fromWindow(self._window)

    return geom

  def _compute_geom(self):
    """Compute size and positions for all windows.

    Returns:    Tuple containing window geometries.
    """
    topgeom = self.get_geom()

    top = topgeom.y
    bottom = topgeom.y + topgeom.h

    geom = Struct()
    gf = GeometryFactory(topgeom)

    geom.errorbox = gf.build_above( h = self._errorbox.lines(topgeom.w),
                                    w = topgeom.w,
                                    y = bottom,
                                    x = topgeom.x)
    bottom -= geom.errorbox.h

    if geom.errorbox.h>0:
      geom.errorline = gf.build_above(h = min(HLine.HEIGHT,bottom),
                                      w = topgeom.w,
                                      y = bottom,
                                      x = topgeom.x)
      bottom -= geom.errorline.h
    else:
      geom.errorline = Geometry(0,0,0,0)


    geom.volume = gf.build_above( h = min(Volume.HEIGHT, bottom),
                                  w = Volume.WIDTH,
                                  y = bottom,
                                  x = topgeom.w-Volume.WIDTH)
    geom.status = gf.build_above( h = min(Status.HEIGHT, bottom),
                                  w = topgeom.w-geom.volume.w,
                                  y = bottom,
                                  x = topgeom.x)
    bottom -= max(geom.volume.h, geom.status.h)


    geom.controller = gf.build_above( h = min(Controller.HEIGHT, bottom),
                                      w = topgeom.w,
                                      y = bottom,
                                      x = topgeom.x)
    bottom -= geom.controller.h


    geom.about = gf.build(h = min(About.HEIGHT,bottom),
                          w = About.WIDTH,
                          y = top,
                          x = topgeom.w-About.WIDTH)
    geom.dirname = gf.build(h = min(DirName.HEIGHT,bottom),
                            w = geom.about.x,
                            y = top,
                            x = topgeom.x)
    top += max(geom.about.h, geom.dirname.h)

    geom.selectline = gf.build( h = min(HLine.HEIGHT,bottom-top),
                                w = topgeom.w,
                                y = top,
                                x = topgeom.x)
    top += geom.selectline.h


    geom.select = gf.build( h = bottom-top,
                            w = topgeom.w,
                            y = top,
                            x = topgeom.x)

    return geom


  def _on_event(self,obj,event):
    """Handle user input events.

    Arguments:
      obj       The controller this event came from.
      event     The string containing the command to evaluate.

    Returns:    True if the event was handled.
    """
    data = event.split()
    if len(data)>0:
      cmd = data[0]
      if cmd=="resize":
        self._update_layout()
        return True
      elif cmd=="redraw":
        self._errorbox.update()
        self._errorline.update()
        self._controller.update()
        self._volume.update()
        self._status.update()
        self._select.update()
        self._dirname.update()
        self._about.update()
        self._selectline.update()
        return True
      elif cmd=="quit":
        self.quit()
        return True

      return False
          

  def _update_layout(self):
    """Update the window layout.
    """
    try:
      geom = self._compute_geom()
      self._errorbox.set_geom(geom.errorbox)
      self._errorline.set_geom(geom.errorline)
      self._controller.set_geom(geom.controller)
      self._volume.set_geom(geom.volume)
      self._status.set_geom(geom.status)
      self._select.set_geom(geom.select)
      self._dirname.set_geom(geom.dirname)
      self._about.set_geom(geom.about)
      self._selectline.set_geom(geom.selectline)
    except curses.error:
      pass

  def quit(self):
    """Shut down the audiobook player.
    """
    self._fiforeader.quit()
    self._audiobook.pause()
    self._audiobook.quit()
    self._quit = True

  def run(self,filename,position):
    """Run the audiobook player.
    
    Arguments:
      filename  The filename to start playing at. (Or None to let the player
                decide.)

      position  The position to start playing at in nanoseconds. (Or none to
                let the player decide.)
    """
    # Set default volume.
    volume_control = self._audiobook.gst().get_by_name('volume-control')
    if volume_control != None:
      volume_control.set_property("volume",self._conf.volume)

    try:
      # Seek to start position.
      if(self._conf.autoplay):
        self._audiobook.play(filename,position)
      else:
        self._audiobook.seek(filename,position)
      self._loop()
    except SystemExit:
      self.quit()
  
  def _loop(self):
    """Custom glib mainloop that checks for ncurses input.
    """
    # Loop until the program should be shut down.
    while not self._quit:
      # Attempt to handle an glib event (block until something happens.)
      self._gcontext.iteration(True)

      # Poll for ncurses input/resize event.
      #
      # This code is reached after each glib event has been handled.
      # There is a dummy event handler that listens for data on stdin, so this
      # code is reached when there is data on stdin.
      #
      # When the terminal is resized the program receives a signal which causes
      # the method above to return even though no event has been handled. The
      # terminal resize event is handled by the same code as user input due to
      # how ncurses is designed.
      #
      self._controller.poll()

class Actuator(object):
  """A command actuator for the audiobook player.
  """

  def __init__(self,conf,audiobook,controller):
    """Create the actuator

    Arguments:
      conf        Configuration struct. (Config file and arguments.)
      audiobook   The audiobook player object to control.
      controller  The controller that emits events.
    """
    self._audiobook = audiobook
    self._controller = controller
    self._controller.connect("event",self._on_event)

    self._loop = conf.loop
    self._autoquit = conf.autoquit
    self._audiobook.connect("notify::eob",self._on_eob)

  def _on_eob(self,ab,prop):
    """The end of the audiobook has been reached.

    Arguments:
      ab    The audiobook that this is a view for.
      prop  The property that was updated.
    """
    # Make sure we are still at the end.
    if ab.eob:
      if self._loop==0:
        # We are no longer looping.
        if self._autoquit:
          self._controller.emit("event","quit")
      else:
        # We are still looping.
        if self._loop>0:
          self._loop -= 1

        # Start playing at the beginning.
        filelist = ab.list_files()
        if len(filelist)>0:
          ab.play(start_file=filelist[0],start_pos=0)

    return True

  def _on_event(self,obj,event):
    """Handle user input events.

    Arguments:
      obj       The controller this event came from.
      event     The string containing the command to evaluate.

    Returns:    True if the event was handled.
    """
    ab = self._audiobook
    data = event.split()
    if len(data)>0:
      try:
        cmd = data[0]

        if cmd=="play":
          (filename,_,_) = self._audiobook.position()
          if filename != None:
            start_file = self._get_file(data)
            (rel, start_pos) = self._get_pos(data)
            ab.play(start_file=start_file,start_pos=start_pos)
          return True

        elif cmd=="jumpfile" and len(data)==2:
          pos = int(data[1])-1
          filelist = ab.list_files()

          if len(filelist)>0:
            if pos < 0:
              pos = 0
            elif pos >= len(filelist):
              pos = len(filelist)-1

            ab.seek(start_file=filelist[pos],start_pos=0)
          return True

        elif cmd=="pause":
          ab.pause()
          return True

        elif cmd=="seek":
          start_file = self._get_file(data)
          (rel, start_pos) = self._get_pos(data)
          if start_file == None and rel:
            ab.dseek(start_pos)
          else:
            ab.seek(start_file=start_file,start_pos=start_pos)
          return True

        elif cmd=="dseek" and len(data)==2:
          (rel, start_pos) = self._get_pos(data)
          if start_pos != None:
            ab.dseek(start_pos)
          return True

        elif cmd=="stepfile" and len(data)==2:
          delta = int(data[1])
          new_file = ab.get_file(delta)
          if new_file!=None:
            ab.seek(start_file=new_file,start_pos=0)
          return True

        elif cmd=="play_pause" and len(data)==1:
          (filename,_,_) = self._audiobook.position()
          if filename != None:
            ab.play_pause()
          return True

        elif (cmd=="volume" or cmd=="dvolume") and len(data)==2:
          volume_control = self._audiobook.gst().get_by_name('volume-control')
          if volume_control != None:
            raw = data[1]
            if raw[0] == "+" or raw[0] == "-" or cmd == "dvolume":
              oldvol = volume_control.get_property("volume")
              volume = oldvol + float(raw)
            else:
              volume = float(raw)

            volume = max(0, min(volume, 10))
            volume_control.set_property("volume",volume)
          return True

        elif cmd=="mark" and len(data)>=2:
          name = ' '.join(data[1:])
          self._audiobook.mark(name)
          return True

      except ValueError as e:
        pass
    return False

  def _get_file(self,data):
    """Parse a filename from given data.
    
    Arguments:
      data    List of strings representing each word.

    Returns:  Filename, or None if no filename was given.
    """
    if len(data)>=3:
      return " ".join(data[1:-1])
    else:
      return None

  def _get_pos(self,data):
    """Parse position.
    
    Arguments:
      data    List of strings representing each word.

    Returns:  Position, None if no position was given.

    Exceptions:
      ValueError if parsing failed.
    """
    if len(data)>=2:
      (rel, pos) = parse_pos(data[-1])
      if pos == None:
        raise ValueError()
      return (rel, pos)
    else:
      return (None,None)

def run():
  """Run the audiobook curses UI.
  """
  parser = pstorytime.audiobookargs.ArgumentParser(
    add_help=True,
    parents=[pstorytime.audiobookargs.audiobookargs],
    fromfile_prefix_chars="@",
    conflict_handler='resolve',
    formatter_class=argparse.RawDescriptionHelpFormatter,

    description=textwrap.dedent('''\
        %(prog)s is a logging console audiobook player based on gstreamer with
        a curses interface. It stores all events like play, pause and seek
        together with walltime, filename and position in file. It also
        autosaves the position while playing to recover from crashes without
        loosing the position. This allows the user to retrace his/her steps and
        minimizes the risk of getting lost.

        Furthermore, it treats all files as a continuous stream so that seeking
        can be made seamlessly between different files.
        '''),

    epilog=textwrap.dedent('''\
        If -- is used as an argument, no more optional arguments are parsed.
        The rest are expected to be positional.

        Paths can contain the strings {conf} and {audiobook}. They are replaced
        with the absolute path to the configuration directory and the audiobook
        directory respectively. The tilde character is expanded as usual to the
        users home directory. Arguments can be placed one on each line in the
        main configuration file or in any file specified on the commandline
        prefixed by an @ sign. Though {conf} and {audiobook} are not expanded
        in paths preceded by the @ sign. Since there is only one argument on
        each line, no citation marks are needed to escape multi-word strings. 

        Available events (and default bindings):
          up                Move one line up. (KEY_UP, k)
          down              Move one line down. (KEY_DOWN, j)
          ppage             Move one page up, same as up when in error log.
                            (KEY_PPAGE)
          npage             Move one page down, same as down when in error log.
                            (KEY_NPAGE)
          begin             Go to first line of list. (KEY_HOME, H)
          end               Go to end of list. (KEY_END, F)
          next_view         Change to next view. (^I = Tab)
          playlog_view      Change to playlog view. (p)
          filelist_view     Change to filelist view. (f)
          errorlog_view     Change to errorlog view. (e)
          select [POS]      Jump to the currently selected entry. Optionally
                            with an absolute or relative position to jump to in
                            regard to the given file/log entry. (^J = Return:
                            select {b})
          timer [TIME]      Pause playback the next time a certain time of day
                            is reached. The time can be either absolute or
                            relative just like file positions. If no time is
                            given, the timer is stopped. (t: timer {b}; play,
                            T: timer)
          buffer store STR  Add the given string to the input buffer. (0-9, :, +, -)
          buffer erase      Erase a character from the input buffer.
                            (KEY_BACKSPACE, ^?)
          buffer clear      Clear the input buffer. (=)
          seek POS          Seek to an absolute or relative position.
                            (KEY_LEFT, h: seek -10, KEY_RIGHT, l: seek +10)
          dseek POS         Seek to a relative position. Absolute positions are
                            also treated as relative positions.
          stepfile INT      Jump to a different file. The integer indicates
                            an index relative to the current file. For example,
                            -1, 0 and 1 indicates the previous, the current and
                            the next file respectively.
          jumpfile INT      Seek to the beginning of the specified file,
                            indexed from 1 and up.
          volume FLOAT      Change volume. An absolute value is an float
                            between 0.0 and 10.0. The number can be prepended
                            with a + or - to change the volume relative to the
                            current one. (u: volume +0.1, d: volume -0.1)
          dvolume FLOAT     Relatively change the volume. Same as "volume" but
                            it is always treated as a relative change.
          play              Start playback. (t: timer {b}; play)
          pause             Pause playback.
          play_pause        Toggle play/pause. (" " = Space)
          redraw            Force redraw of the interface. (^L)
          resize            Recompute layout and redraw.
          mark NAME         Add a bookmark with the given name. (*: mark *)
          quit              Quit pstorytime. (q, ^C)

        Events may contain the string {b} which is replaced with the current
        contents of the input buffer. Multiple events can be sequenced by
        separating them by a semi-colon and, if entered through the command
        pipe, by placing them on separate lines. When the events are sequenced
        by a semi colon and an event fails, the rest of the events on that line
        will be ignored.

        A position can optionally start or end with a + or - which indicates
        that it is a relative position. This is followed or preceded by a colon
        separated list of integers given as HOURS:MINUTES:SECONDS, where some
        entries may be omitted. Note that the relative position three minutes
        can be given in a number of different ways, including +3:00, +3:0,
        +180, 1:120+, 0:3:0+ and 0:03:00+.
        '''))

  parser.add_argument(
    "--cmdpipe",
    metavar="PATH",
    nargs='?',
    help="Optional path to a pipe that commands are read from relative to current directory. If no path is given, the cmdpipe is disabled. See section on paths. (Default: %(default)s)",
    const=None,
    default="{conf}/cmdpipe")

  parser.add_argument(
    "--playlog-file",
    metavar="PATH",
    help="Path to the file to save playlog in relative to current directory. See section on paths. (Default: %(default)s)",
    default="{conf}/logs/{audiobook}/.playlog")

  parser.add_argument(
    "--conf-dir",
    metavar="PATH",
    help="Configuration directory (Default: %(default)s)",
    default="~/.pstorytime")

  parser.add_argument(
    "--conf",
    metavar="PATH",
    nargs='?',
    help="Path to main configuration file relative to current directory. If no path is given, the main configuration file is disabled. This option has no effect if placed in the main configuration file. See section on paths. (Default: %(default)s)",
    const=None,
    default="{conf}/config")

  parser.add_argument(
    "--lastdir-file",
    metavar="PATH",
    nargs='?',
    help="Path to the file where the last loaded directory is stored. If no path is given, the use of this file is disabled and the current directory will always be used instead. See section on paths. Using the {audiobook} does not make much sense since it is not yet known. (Default: %(default)s)",
    const=None,
    default="{conf}/lastdir")

  parser.add_argument(
    "path",
    help="Audiobook directory, possibly including a file to start playing at. (Default: Load last directory.)",
    nargs='?',
    default=None)

  parser.add_argument(
    "position",
    help="Position to start playing at. See section on positions. (Default: %(default)s)",
    nargs='?',
    action=pstorytime.audiobookargs.Position)

  parser.add_argument(
    "--autoplay",
    metavar="BOOL",
    help="Start playing when the audiobook is started. (Default: %(default)s)",
    action=pstorytime.audiobookargs.Boolean,
    default=False)

  parser.add_argument(
    "--autoquit",
    metavar="BOOL",
    help="Quit when the audiobook ends and the player is not looping any more. (Default: %(default)s)",
    action=pstorytime.audiobookargs.Boolean,
    default=False)

  parser.add_argument(
    "--loop",
    metavar='N',
    help="Number of times to continue playing from the beginning when the audiobook ends. Negative number for infinitely many times. (Default: %(default)s)",
    default=0,
    type=int)

  parser.add_argument(
    "--volume",
    metavar="VOL",
    help="Playback volume as a float between 0 and 10. (Default: %(default)s)",
    default=1.0,
    type=float)

  parser.add_argument(
    "--default-bindings",
    metavar="BOOL",
    help="Load default bindings. Otherwise all bindings need to be added manually. (Default: %(default)s)",
    action=pstorytime.audiobookargs.Boolean,
    default=True)

  parser.add_argument(
    "--bind",
    metavar=('KEY','EVENT'),
    help="Add new binding from key to event. Key names are displayed in the program when pressed, possible events are listed in its own section.",
    nargs=2,
    action='append',
    default=[])

  parser.add_argument(
    "--unbind",
    metavar='KEY',
    help="Remove binding for the given key. Key names are displayed in the program when pressed.",
    nargs=1,
    action='append',
    dest='bind',
    default=[])

  parser.add_argument(
    "--event-len",
    metavar='N',
    help="Number of characters to display of event names in the playlog. (Default: %(default)s)",
    default=8,
    type=int)

  parser.add_argument(
    "--keytime",
    metavar='MSECS',
    help="Number of milliseconds to display the name of the last key pressed. (Default: %(default)s)",
    default=1000,
    type=int)

  parser.add_argument(
    "--errortime",
    metavar='MSECS',
    help="Number of milliseconds to display the last error message, set it to a negative value to show it forever. (Default: %(default)s)",
    default=5000,
    type=int)

  parser.add_argument(
    "-v", "--version",
    help="Show version information.",
    version='%(prog)s {0}'.format(pstorytime.version.version),
    action='version')

  parser.add_argument(
    "--banner",
    help="Show pandora rebirth competition banner.",
    action='store_true')

  parser.add_argument(
    "--gst-pipeline",
    metavar='STR',
    help="The gstreamer pipeline to use. Should contain a filesrc and volume component named file-source and volume-control respectively. (Default: %(default)s)",
    default='filesrc name=file-source ! decodebin2 name=decoder identity name=postdecode ! audioconvert ! audioresample ! volume name=volume-control ! autoaudiosink')

  parser.add_argument(
    "--gst-link",
    metavar=('FROM','TO'),
    help="Automatically add links between given pairs of components if possible. (Default: decoder -> postdecode)",
    nargs=2,
    action='append',
    default=[('decoder','postdecode')])

  # argparse crashes if there is an empty string in sys.argv.
  if "" in sys.argv:
    print("Empty strings are not allowed as arguments.")
    exit(1)

  # Allow tilde in @-paths.
  for i in xrange(0,len(sys.argv)):
    if sys.argv[i][0] == "@":
      sys.argv[i] = "@" + expanduser(sys.argv[i][1:])

  # First pass of reading arguments.
  conf = parser.parse_args()

  # If a config file is given.
  if conf.conf != None and conf.conf != "":
    # Find out preliminary audiobook directory.
    if conf.path != None:
      if isdir(conf.path):
        directory = conf.path
      else:
        directory = dirname(conf.path)
    else:
      directory = ""

    # Find out full path to config file.
    gen = PathGen(conf.conf_dir,directory)
    configfile = gen.gen(conf.conf)

    # If configuration file is found:
    if isfile(configfile):
      # Read it, and then read the arguments again.
      conf = parser.parse_args(["@"+configfile])
      conf = parser.parse_args(namespace=conf)

  gen = PathGen(conf.conf_dir,"")
  conf.lastdir_file = gen.gen(conf.lastdir_file)
  lastdir = LineStore(conf.lastdir_file,".")

  if conf.banner:
    pstorytime.pandorabanner.show()
    sys.exit(0)

  if conf.path == None:
    # No path was given, use the last directory that was played from.
    # If the file for storing this is deleted, use the current directory.
    directory = lastdir.get()
    filename = None
    if directory == None:
      print("No last audiobook directory available.")
      exit(1)
  elif isfile(conf.path):
    # The path points to a file, start playing this specific file, ignoring the
    # last playlog entry.
    directory = dirname(conf.path)
    filename = basename(conf.path)
  elif isdir(conf.path):
    # The path points to a directory, use the last playlog entry to find out
    # where to start playing.
    directory = conf.path
    filename = None
  else:
    print("No such file or directory: {0}".format(conf.path))
    exit(1)

  directory = abspath(expanduser(directory))

  if not isdir(directory):
    print("No such directory: {0}".format(directory))
    exit(1)

  # Update the file containing the last directory loaded.
  lastdir.set(directory)

  gen = PathGen(conf.conf_dir,directory)
  conf.playlog_file = gen.gen(conf.playlog_file)
  conf.cmdpipe = gen.gen(conf.cmdpipe)

  if conf.cmdpipe==None or conf.cmdpipe=="":
    pipelock = DummyLock()
  else:
    pipelock = FileLock(conf.cmdpipe+".lock")

  dirlock = FileLock(conf.playlog_file+".lock")

  try:
    # Make sure there are no other instances running that is using the same files.
    with pipelock:
      with dirlock:
        def run(stdscr):
          """Launch curses UI.

          Arguments:
            stdscr      Main ncurses window.
          """
          ui = CursesUI(conf,directory,stdscr)
          ui.run(filename,conf.position)
        # Set up the terminal for curses use.
        curses.wrapper(run)
  except LockedException:
    print("Error: Another instance is already using the same files.")
    sys.exit(1)
  except gobject.GError as e:
    print("GStreamer error: {0}".format(e))
    sys.exit(1)

if __name__ == '__main__':
  run()
